from wacky_rl.losses._loss_core import WackyLoss

from wacky_rl.losses.actor_losses import ActorLoss
from wacky_rl.losses.actor_losses import SoftActorLoss
from wacky_rl.losses.actor_losses import PPOActorLoss

from wacky_rl.losses.q_losses import SoftQLoss

from wacky_rl.losses.val_losses import SoftValueLoss
from wacky_rl.losses.val_losses import MeanSquaredErrorLoss