#from wacky_rl.layers.actor_layer import DiscreteActionLayer
#from wacky_rl.layers.actor_layer import ContinActionLayer

from wacky_rl.layers.contin_actor_layer import ContinActionLayer
from wacky_rl.layers.discrete_actor_layer import DiscreteActionLayer